unit rkSmartPath;

{$WARN SYMBOL_PLATFORM OFF}
{$WARN UNIT_PLATFORM OFF}

interface

uses
  Windows, SysUtils, Classes, Controls, ExtCtrls, Graphics, Forms, Messages,
  ShellAPI, ImgList, Dialogs, Menus, StdCtrls;

const
  CM_REFRESH = WM_USER + 1295; // Custom Message...

type
  TspItemState = (spNormal, spButtonPressed, spArrowPressed, spButtonHot,
    spArrowHot);
  TspPaths = set of (spDesktop, spDocuments, spPictures, spMusic, spVideos);
  TOnOwnerDrawEvent = procedure(Sender: TObject; ACanvas: TCanvas; ARect: TRect;
    AText: string; AState: TspItemState; AButton: Boolean) of object;
  TOnTextEvent = procedure(Sender: TObject; AText: string) of object;

  TrkSmartPath = class(TCustomControl)
  private
    { Private declarations }
    FAuto: Boolean;
    FAllowEdit: Boolean;
    FBmp: TBitmap;
    FBorderColor: TColor;
    FBorderStyle: TBorderStyle;
    FClicked: Boolean;
    FColor: TColor;
    FCustomPaths: TStringList;
    FCustomPopup: TPopupMenu;
    FDefault: Boolean;
    FDesktopFolder: string;
    FEmptyPathIcon: Integer;
    FEmptyPathText: string;
    FFont: TFont;
    FFramed: Boolean;
    FHotPt: TPoint;
    FHotPos: Integer;
    FImages: TImageList;
    FMenu: TPopupMenu;
    FMenuLookup: TStringList;
    FNewFolder: string;
    FPath: string;
    FPathEdit: TEdit;
    FPathIdx: Integer;
    FPaths: TStringList;
    FPathsShow: TStringList;
    FPopupPath: string;
    FPopUp: Boolean;
    FShowArrow: Boolean;
    FSpecialFolders: TspPaths;
    FState: TspItemState;
    FOnOwnerDraw: TOnOwnerDrawEvent;
    FOnPathChanged: TNotifyEvent;
    FOnSearch: TOnTExtEvent;
    FAllowSearch: Boolean;
    procedure WMERASEBKGND(var Message: TWMEraseBkgnd); message WM_ERASEBKGND;
    procedure WMPaint(var Message: TWMPaint); message WM_PAINT;
    procedure CMCtl3DChanged(var Message: TMessage); message CM_CTL3DCHANGED;
    procedure WMWindowPosChanged(var Message: TWMWindowPosChanged); message
      WM_WINDOWPOSCHANGED;
    procedure SetBorderStyle(Value: TBorderStyle);
    procedure SetFont(Value: TFont);
    procedure PaintPath;
    procedure SetPath(Value: string);
    procedure SetCustomPaths(Value: TStringList);
    procedure SetPathsToAdd(Value: TspPaths);
    function CalcPath: Integer;
    procedure CMRefresh(var Message: TMessage); message CM_REFRESH;
    procedure CalcState;
    procedure EditExit(Sender: TObject);
    procedure EditKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure EditKeyPress(Sender: TObject; var Key: Char);
    procedure SetBorderColor(const Value: TColor);
    procedure SetColor(const Value: TColor);
    procedure SetFramed(const Value: Boolean);
    procedure SetDefault(const Value: Boolean);
  protected
    { Protected declarations }
    FPopupPos: TPoint;
    FPathOff: Integer;
    FImageIdx: Integer;
    FCreateFolder: Boolean;
    procedure FMenuPopup(Sender: TObject);
    procedure FMenuClick(Sender: TObject);
    procedure PaintWindow(DC: HDC); override;
    procedure Resize; override;
    procedure CreateParams(var Params: TCreateParams); override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y:
      Integer);
      override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
      override;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    procedure CMMouseLeave(var Message: TMessage); message CM_MOUSELEAVE;
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function GetSpecialFolderPath(folder: Integer): string;
    function SysImageListHandle(const Path: string; const WantLargeIcons:
      Boolean): Windows.THandle;
    procedure DoEdit(CreateFolder: Boolean);
    procedure ExitEditMode;
    procedure UpdateView;
    property ImageIdx: Integer read FImageIdx;
    property PopupPath: string read FPopupPath;
    property PathOff: Integer read FPathOff;
    property PathParts: TStringList read FPaths;
    property PathDisplay: TStringList read FPathsShow;
  published
    { Published declarations }
    property Align;
    property AllowEdit: Boolean read FAllowEdit write FAllowEdit default False;
    property AllowSearch: Boolean read FAllowSearch write FAllowSearch default False;
    property Anchors;
    property Auto: Boolean read FAuto write FAuto default True;
    property AutoSize;
    property BevelInner;
    property BevelOuter;
    property BevelKind;
    property BevelWidth;
    property BorderColor: TColor read FBorderColor write SetBorderColor default
      clSilver;
    property BorderStyle: TBorderStyle read FBorderStyle write SetBorderStyle
      default bsNone;
    property BorderWidth;
    property Color: TColor read FColor write SetColor default clWindow;
    property CustomPopupMenu: TPopupMenu read FCustomPopup write FCustomPopup;
    property ComputerAsDefault: Boolean read FDefault write SetDefault;
    property Enabled;
    property EmptyPathIcon: Integer read FEmptyPathIcon write FEmptyPathIcon;
    property EmptyPathText: string read FEmptyPathText write FEmptyPathText;
    property Favorites: TStringList read FCustomPaths write SetCustomPaths;
    property Font: TFont read FFont write SetFont;
    property Framed: Boolean read FFramed write SetFramed default True;
    property Height;
    property Images: TImageList read FImages;
    property Left;
    property NewFolderName: string read FNewFolder write FNewFolder;
    property ParentBackground;
    property Path: string read FPath write SetPath;
    property PopupMenu;
    property ShowHint;
    property SpecialFolders: TspPaths read FSpecialFolders write SetPathsToAdd;
    property TabOrder;
    property TabStop;
    property Top;
    property Visible;
    property Width;
    property OnClick;
    property OnDblClick;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyUp;
    property OnMouseEnter;
    property OnMouseLeave;
    property OnMouseDown;
    property OnMouseUp;
    property OnMouseMove;
    property OnOwnerDraw: TOnOwnerDrawEvent read FOnOwnerDraw write
      FOnOwnerDraw;
    property OnPathChanged: TNotifyEvent read FOnPathChanged write
      FOnPathChanged;
    property OnResize;
    property OnSearch: TOnTExtEvent read FOnSearch write FOnSearch;
  end;

procedure PaintArrow(Canvas: TCanvas; x, y: Integer; Down: Boolean);

procedure Register;

implementation

uses
  ShlObj, SHFolder;

function PathToPIDL(APath: WideString): PItemIDList;
// Takes the passed Path and attempts to convert it to the equavalent PIDL
var
  Desktop: IShellFolder;
  pchEaten, dwAttributes: ULONG;
begin
  Result := nil;
  SHGetDesktopFolder(Desktop);
  dwAttributes := 0;
  if Assigned(Desktop) then
  begin
    if Desktop.ParseDisplayName(0, nil, PWideChar(APath), pchEaten, Result,
      dwAttributes) <> NOERROR then
      Result := nil
  end
end;

function TrkSmartPath.GetSpecialFolderPath(folder: Integer): string;
const
  SHGFP_TYPE_CURRENT = 0;
var
  path: array[0..MAX_PATH] of char;
begin
  if SUCCEEDED(SHGetFolderPath(0, folder, 0, SHGFP_TYPE_CURRENT, @path[0])) then
    Result := path
  else
    Result := '';
end;

// Gets handle to system image list for a specified path

function TrkSmartPath.SysImageListHandle(const Path: string; const
  WantLargeIcons: Boolean): Windows.THandle;
var
  FI: ShellAPI.TSHFileInfo;
  Flags: Windows.UINT;
begin
  Flags := ShellAPI.SHGFI_SYSICONINDEX;
  if WantLargeIcons then
    Flags := Flags or ShellAPI.SHGFI_LARGEICON
  else
    Flags := Flags or ShellAPI.SHGFI_SMALLICON;
  Result := ShellAPI.SHGetFileInfo(PChar(Path), 0, FI, SizeOf(FI), Flags);
end;

procedure TrkSmartPath.UpdateView;
begin
  PaintPath;
end;

function AddBackSlash(const AString: string): string;
begin
  if (AString = '') then
    Result := ''
  else
  begin
    if (AString[Length(AString)] <> '\') then
      Result := AString + '\'
    else
      Result := AString;
  end;
end;

// returns a unique folder name for this path, path needs a backslash

function GetNewFolderName(const Path, FolderName: string): string;
var
  i: Integer;
begin
  Result := Path + FolderName;
  i := 0;
  while DirectoryExists(result) do
  begin
    inc(i);
    Result := Path + FolderName + ' ' + IntToStr(i);
  end;
end;

procedure TrkSmartPath.ExitEditMode;
begin
  if FPathEdit.Visible then
  begin
    FPathEdit.Visible := False;
    Invalidate;
    SetFocus;
  end;
end;

procedure TrkSmartPath.DoEdit(CreateFolder: Boolean);
var
  i: Integer;
  txt: string;
begin
  FCreateFolder := CreateFolder;
  FPathEdit.Color := FColor;
  FPathEdit.Left := 4;
  FPathEdit.Width := ClientWidth - 8;
  FPathEdit.Font.Assign(FFont);
  FPathEdit.Height := FPathEdit.Font.Size + 6;
  FPathEdit.Top := ((ClientHeight - FPathEdit.Height) div 2);
  FPathEdit.AutoSelect := not CreateFolder;
  if CreateFolder then
  begin
    txt :=AddBackSlash(Path);
    i := Length(txt);
    txt := GetNewFolderName(txt, FNewFolder);
    FPathEdit.Text := txt;
    FPathEdit.SelStart := i;
    FPathEdit.SelLength := Length(txt) - i;
  end
  else
    FPathEdit.Text := Path;
  FPathEdit.Visible := True;
  PaintPath;
  FPathEdit.SetFocus;
end;

procedure TrkSmartPath.EditKeyDown(Sender: TObject; var Key: Word; Shift:
  TShiftState);
var
  done: Boolean;
begin
  if key = VK_Escape then
    done := True
  else if key = VK_Return then
  begin
    done := True;
    if FCreateFolder then
    begin
      if not CreateDir(FPathEdit.Text) then
        ShowMessage('Create folder failed. An error occured. ErrorCode: ' +
          IntToStr(GetLastError));
    end;

    if (Pos(':', FPathEdit.Text) <> 0) then
      Path := FPathEdit.Text
    else
    begin
      if FAllowSearch then
        if Assigned(FOnSearch) then
          FOnSearch(Self, FPathEdit.Text);
      done:= False;
    end;
    {
    if (FCLIMode) and Assigned(FOnCliCommand) then
    begin
      FOnCliCommand(Self, FPathEdit.Text);
      done := False;
    end;
    if Trim(UpperCase(FPathEdit.Text)) = 'CLI' then
    begin
      done := False;
      FCLIMode := True;
      FPathEdit.Text := '';
      FPathEdit.Color := clBlack;
      FPathEdit.Font.Color := clWhite;
      Invalidate;
    end;

    if done then
      Path := FPathEdit.Text;
    }
  end
  else
    done := False;
  if done then
    ExitEditMode;
end;

// to avoid the beeping when pressing enter and esc key

procedure TrkSmartPath.EditKeyPress(Sender: TObject; var Key: Char);
begin
  if (Key = #13) or (Key = #27) then
    Key := #0;
end;

procedure TrkSmartPath.EditExit(Sender: TObject);
begin
  ExitEditMode;
end;

procedure MultiSzToStrings(const MultiSz: PChar; const Strings:
  Classes.TStrings);
var
  P: PChar;
begin
  if not Assigned(MultiSz) then
    Exit;
  P := MultiSz;
  while P^ <> #0 do
  begin
    Strings.Add(P);
    Inc(P, SysUtils.StrLen(P) + 1);
  end;
end;

procedure TrkSmartPath.FMenuPopup(Sender: TObject);
const
  Flag: Cardinal = SHGFI_SMALLICON or SHGFI_SYSICONINDEX or
  SHGFI_PIDL or SHGFI_TYPENAME or SHGFI_DISPLAYNAME;
  SSI: Integer = SizeOf(TSHFileInfo);
var
  AMenuItem: TMenuItem;
  SFI: TSHFileInfo;
  SR: TSearchRec;
  IsFound: Boolean;
  i, j, idx: Integer;
  FName, s: string;
  Drives: PChar;
  BufSize: Integer;
  List: TStringList;
  SelectedPIDL: PItemIDList;

  procedure AddMenuItem(PName, FName: string; Icon: Integer);
  begin
    AMenuItem := TMenuItem.Create(FMenu);
    AMenuItem.Caption := FName;
    AMenuItem.OnClick := FMenuClick;
    AMenuItem.ImageIndex := Icon;
    AmenuItem.Tag := Idx;
    Inc(idx);
    FMenu.Items.Add(AMenuItem);
    FMenuLookup.Add(PName);
  end;

begin
  Screen.Cursor := crHourglass;
  for i := FMenu.Items.Count - 1 downto 0 do
    FMenu.Items[i].Free;
  FMenuLookup.Clear;
  idx := 0;
  if FPopupPath = 'HomeFolder' then
  begin
    if FPathOff > 0 then
    begin
      for i := FPathOff - 1 downto 0 do
      begin
        FName := '';
        for j := 0 to i do
          FName := FName + FPaths[j] + '\';

        SelectedPIDL := PathToPIDL(FName);
        SHGetFileInfo(PChar(SelectedPIDL), 0, SFI, SSI, Flag);
        AddMenuItem(FName, SFI.szDisplayName, SFI.iIcon);
      end;
      AddMenuItem('- ', '-', -1);
    end;
    AddMenuItem('*', 'New Folder', -1);
    AddMenuItem('- ', '-', -1);
    if spDesktop in FSpecialFolders then
    begin
      FName := GetSpecialFolderPath(CSIDL_DESKTOPDIRECTORY);
      s := ExtractFileName(Fname);

      SelectedPIDL := PathToPIDL(FName);
      SHGetFileInfo(PChar(SelectedPIDL), FILE_ATTRIBUTE_NORMAL, SFI, SSI, Flag);
      AddMenuItem(FName, SFI.szDisplayName, SFI.iIcon);
    end;
    if spDocuments in FSpecialFolders then
    begin
      FName := GetSpecialFolderPath(CSIDL_PERSONAL);
      s := ExtractFileName(Fname);
      SelectedPIDL := PathToPIDL(FName);
      SHGetFileInfo(PChar(SelectedPIDL), FILE_ATTRIBUTE_NORMAL, SFI, SSI, Flag);
      AddMenuItem(FName, SFI.szDisplayName, SFI.iIcon);
    end;
    if spPictures in FSpecialFolders then
    begin
      FName := GetSpecialFolderPath(CSIDL_MYPICTURES);
      s := ExtractFileName(Fname);
      SelectedPIDL := PathToPIDL(FName);
      SHGetFileInfo(PChar(SelectedPIDL), FILE_ATTRIBUTE_NORMAL, SFI, SSI, Flag);
      AddMenuItem(FName, SFI.szDisplayName, SFI.iIcon);
    end;
    if spMusic in FSpecialFolders then
    begin
      FName := GetSpecialFolderPath(CSIDL_MYMUSIC);
      s := ExtractFileName(Fname);
      SelectedPIDL := PathToPIDL(FName);
      SHGetFileInfo(PChar(SelectedPIDL), FILE_ATTRIBUTE_NORMAL, SFI, SSI, Flag);
      AddMenuItem(FName, SFI.szDisplayName, SFI.iIcon);
    end;
    if spVideos in FSpecialFolders then
    begin
      FName := GetSpecialFolderPath(CSIDL_MYVIDEO);
      s := ExtractFileName(Fname);
      SelectedPIDL := PathToPIDL(FName);
      SHGetFileInfo(PChar(SelectedPIDL), FILE_ATTRIBUTE_NORMAL, SFI, SSI, Flag);
      AddMenuItem(FName, SFI.szDisplayName, SFI.iIcon);
    end;
    AddMenuItem('- ', '-', -1);
    for i := 0 to FCustomPaths.Count - 1 do
    begin
      FName := FCustomPaths[i];
      s := ExtractFileName(Fname);
      SelectedPIDL := PathToPIDL(FName);
      SHGetFileInfo(PChar(SelectedPIDL), FILE_ATTRIBUTE_NORMAL, SFI, SSI, Flag);
      AddMenuItem(FName, SFI.szDisplayName, SFI.iIcon);
    end;
    AddMenuItem('- ', '-', -1);
    List := TStringList.Create;
    BufSize := Windows.GetLogicalDriveStrings(0, nil);
    GetMem(Drives, BufSize * SizeOf(Char));
    try
      if Windows.GetLogicalDriveStrings(BufSize, Drives) = 0 then
        SysUtils.RaiseLastOSError;
      MultiSzToStrings(Drives, List);
    finally
      FreeMem(Drives);
    end;
    for i := 0 to List.Count - 1 do
    begin
      FName := List[i];
      SelectedPIDL := PathToPIDL(FName);
      SHGetFileInfo(PChar(SelectedPIDL), FILE_ATTRIBUTE_NORMAL, SFI, SSI, Flag);
      AddMenuItem(FName, SFI.szDisplayName, SFI.iIcon);
    end;
    List.Free;
  end
  else
  begin
    IsFound := FindFirst(FPopupPath + '*.*', faAnyFile - faHidden, SR) = 0;
    while IsFound do
    begin
      if (SR.Name <> '.') and (SR.Name <> '..') then
      begin
        if ((SR.Attr and faDirectory) <> 0) then
        begin
          FName := FPopupPath + SR.Name;
          SelectedPIDL := PathToPIDL(FName);
          SHGetFileInfo(PChar(SelectedPIDL), FILE_ATTRIBUTE_NORMAL, SFI, SSI,
            Flag);
          AddMenuItem(SR.Name, SFI.szDisplayName, SFI.iIcon);
        end;
      end;
      IsFound := FindNext(SR) = 0;
    end;
    FindClose(SR);
  end;
  Screen.Cursor := crDefault;
end;

procedure TrkSmartPath.FMenuClick(Sender: TObject);
var
  txt: string;
begin
  if FPopupPath = 'HomeFolder' then
  begin
    txt := FMenuLookup[(Sender as TMenuItem).Tag];
    if txt = '*' then
      DoEdit(True)
    else
      Path := txt
  end
  else
    Path := FPopupPath + FMenuLookup[(Sender as TMenuItem).Tag];
end;

procedure TrkSmartPath.PaintWindow(DC: HDC);
begin
  Canvas.Lock;
  try
    Canvas.Handle := DC;
    try
      FBmp.Canvas.Font.Assign(Font);
      PaintPath;
    finally
      Canvas.Handle := 0;
    end;
  finally
    Canvas.Unlock;
  end;
end;

procedure FillGradient(const Canvas: TCanvas; const ARect: TRect;
  const StartColor, EndColor: TColor);
type
  PRGBTripleArray = ^TRGBTripleArray;
  TRGBTripleArray = array[0..1024] of TRGBTriple;
  TGradientColors = array[0..255] of TRGBTriple;
var
  rc1, gc1, bc1, rc2, gc2, bc2, rc3, gc3, bc3,
    y1, i, GSize: Integer;
  Row: PRGBTripleArray;
  GradCol: TRGBTriple;
  GradientBmp: TBitmap;
begin
  rc1 := GetRValue(ColorToRGB(StartColor));
  gc1 := GetGValue(ColorToRGB(StartColor));
  bc1 := GetBValue(ColorToRGB(StartColor));
  rc2 := GetRValue(ColorToRGB(EndColor));
  gc2 := GetGValue(ColorToRGB(EndColor));
  bc2 := GetBValue(ColorToRGB(EndColor));
  rc3 := rc1 + (((rc2 - rc1) * 15) div 9);
  gc3 := gc1 + (((gc2 - gc1) * 15) div 9);
  bc3 := bc1 + (((bc2 - bc1) * 15) div 9);
  if rc3 < 0 then
    rc3 := 0
  else if rc3 > 255 then
    rc3 := 255;
  if gc3 < 0 then
    gc3 := 0
  else if gc3 > 255 then
    gc3 := 255;
  if bc3 < 0 then
    bc3 := 0
  else if bc3 > 255 then
    bc3 := 255;
  GradientBMP := TBitmap.Create;
  GradientBmp.PixelFormat := pf24bit;
  GradientBmp.Width := 1;
  GradientBmp.Height := (ARect.Bottom - ARect.Top) - 1;
  GSize := GradientBmp.Height;
  y1 := GSize div 2;
  for i := 0 to GSize - 1 do
  begin
    Row := PRGBTripleArray(GradientBmp.ScanLine[i]);
    if (i < y1 - 1) then
    begin
      GradCol.rgbtRed := Byte(rc1 + (((rc2 - rc1) * (i)) div y1));
      GradCol.rgbtGreen := Byte(gc1 + (((gc2 - gc1) * (i)) div y1));
      GradCol.rgbtBlue := Byte(bc1 + (((bc2 - bc1) * (i)) div y1));
    end
    else
    begin
      GradCol.rgbtRed := Byte(rc3 + (((rc2 - rc3) * (i)) div GSize));
      GradCol.rgbtGreen := Byte(gc3 + (((gc2 - gc3) * (i)) div GSize));
      GradCol.rgbtBlue := Byte(bc3 + (((bc2 - bc3) * (i)) div GSize));
    end;
    Row[0] := GradCol;
  end;
  Canvas.StretchDraw(ARect, GradientBmp);
  GradientBmp.Free;
end;

procedure TrkSmartPath.CalcState;
var
  i, n: Integer;
  txt: string;
  R: TRect;
  ShowArrow, done, bool: Boolean;
  popX, popY: Integer;
  s: string;
begin
  FPathOff:= CalcPath;
  R.Left := 22;
  n := FPathOff - 1;
  bool := True;
  done := False;
  ShowArrow := True;
  FState := spNormal;
  s := '';
  txt := '';
  for i := 0 to n do
    s := s + FPaths[i] + '\';
  while not done do
  begin
    FBmp.Canvas.Font.Assign(FFont);
    done := (n = FPaths.Count - 1);
    if bool then
      R.Right := R.Left + 14;
    if not bool then
    begin
      inc(n);
      if (n > -1) and (n < FPaths.Count) then
        txt := FPaths[n];
      s := s + txt + '\';
      R.Right := R.Left + FBmp.Canvas.TextWidth(FPathsShow[n]) + 9;
    end;
    if done then
      ShowArrow := (FShowArrow) or (FPathOff = FPaths.Count);
    if ShowArrow then
    begin
      if (FHotPt.X >= R.Left) and (FHotPt.X <= R.Right) then
      begin
        if FClicked then
        begin
          if bool then
            FState := spArrowPressed
          else
            FState := spButtonPressed;
        end
        else
        begin
          if bool then
            FState := spArrowHot
          else
            FState := spButtonHot;
        end;
        popx := R.Right - 22;
        popy := ClientHeight - 1;
        FPopupPos := ClientToScreen(Point(popX, popY));
        if (bool) and (n = FPathOff - 1) then
          s := 'HomeFolder';
        FPopupPath := s;
        FHotPos := n;
      end;
    end;
    R.Left := R.Right;
    bool := not bool;
  end;
end;

procedure PaintArrow(Canvas: TCanvas; x, y: Integer; Down: Boolean);
begin
  with Canvas do
  begin
    if Down then
    begin
      MoveTo(x + 3, y + 3);
      LineTo(x + 7, y + 3);
      LineTo(x + 5, Y + 5);
      LineTo(x + 3, y + 3);
      LineTo(x + 6, y + 5);
      MoveTo(x + 2, y + 2);
      LineTo(x + 9, y + 2);
    end
    else
    begin
      MoveTo(x + 4, y);
      LineTo(x + 4, y + 4);
      LineTo(x + 6, y + 2);
      LineTo(x + 4, y);
      LineTo(x + 5, y + 3);
      LineTo(x + 3, y - 1);
      LineTo(x + 3, y + 6);
    end;
  end;
end;

procedure PaintButton(ACanvas: TCanvas; ARect: TRect; AText: string;
  AState: TspItemState; AButton: Boolean);
var
  c1, c2: TColor;
  Grey: Boolean;
  r: TRect;
  y: Integer;
begin
  r := ARect;
  if not AButton then
    r.Left := r.Left - 1;
  if AState <> spNormal then
  begin
    ACanvas.Font.Color := clBlack;
    Grey := AButton and (AState = spArrowHot);
    if Grey then
    begin
      c1 := $F2F2F2;
      c2 := $E3E3E3;
    end
    else
    begin
      if (AState = spArrowHot) then
      begin
        c1 := $FDF6EA;
        c2 := $F4D9A6;
      end
      else
      begin
        c1 := $FDF6EA;
        c2 := $FADFAC;
      end;
    end;
    if Grey then
      ACanvas.Pen.Color := $8F8F8F
    else
      ACanvas.Pen.Color := $B17F3C;
    ACanvas.MoveTo(r.Left, r.Top);
    ACanvas.LineTo(r.Left, r.Bottom);
    ACanvas.MoveTo(r.Right - 1, r.Top);
    ACanvas.LineTo(r.Right - 1, r.Bottom);
    r.Left := r.Left + 2;
    r.Right := r.Right - 2;
    r.Top := R.Top + 1;
    r.Bottom := R.Bottom - 1;
    if (AState = spArrowHot) or (AState = spButtonHot) then
      FillGradient(ACanvas, r, c1, c2)
    else
    begin
      InflateRect(r, 1, 1);
      FillGradient(ACanvas, r, c1, c2);
      InflateRect(r, -1, -1);
      r.Top := r.Top - 1;
      r.Bottom := r.Bottom + 1;
      ACanvas.Pen.Color := $927764;
      ACanvas.MoveTo(r.Left - 1, r.Bottom - 1);
      ACanvas.LineTo(r.Left - 1, r.Top);
      ACanvas.LineTo(r.Right + 1, r.Top);
      ACanvas.MoveTo(r.Right, r.Top + 1);
      ACanvas.Pen.Color := $EACF9C;
      ACanvas.LineTo(r.Left, r.Top + 1);
      ACanvas.LineTo(r.Left, r.Bottom - 1);
      ACanvas.LineTo(r.Right + 1, r.Bottom - 1);
    end;
  end;
  if AButton then
  begin
    ACanvas.Brush.Style := bsClear;
    if (AState = spArrowPressed) or (AState = spButtonPressed) then
    begin
      r.Left := r.Left + 1;
      r.Top := r.Top + 1;
      r.Bottom := r.Bottom + 1;
    end;
    DrawText(ACanvas.Handle, PChar(AText), Length(AText), r,
      DT_CENTER or DT_VCENTER or DT_SINGLELINE);
  end
  else
  begin
    y := (((ARect.Bottom - ARect.Top) - 7) div 2) + 2;
    if AState = spNormal then
      ACanvas.Pen.Color := ACanvas.Font.Color
    else
      ACanvas.Pen.Color := clBlack;
    PaintArrow(ACanvas, ARect.Right - 12, y, (AState = spButtonPressed) or
      (AState = spArrowPressed));
  end;
end;

procedure TrkSmartPath.PaintPath;
var
  i, n: Integer;
  txt: string;
  R: TRect;
  ShowArrow, Done, Bool: Boolean;
begin
  CalcState;
  FBmp.Width := ClientWidth;
  FBmp.Height := ClientHeight;
  FBmp.Canvas.Brush.Style := bsSolid;
  FBmp.Canvas.Brush.Color := FColor;
  FBmp.Canvas.Pen.Color := FColor;
  R := ClientRect;
  if FFramed then
    FBmp.Canvas.Pen.Color := FBorderColor;
  FBmp.Canvas.Rectangle(R);
  if FFramed then
    InflateRect(R, -1, -1);
  if not FPathEdit.Visible then
  begin
    i := ((ClientHeight - 16) div 2) + 1;
    if (FImageIdx <> -1) then
      FImages.Draw(FBmp.Canvas, 4, i, FImageIdx);
    R.Left := 22;
    txt := '';
    n := FPathOff - 1;
    bool := True;
    done := False;
    ShowArrow := True;
    while not done do
    begin
      done := (n = FPathsShow.Count - 1);
      if bool then
        R.Right := R.Left + 14;
      if not bool then
      begin
        inc(n);
        if (n > -1) and (n < FPathsShow.Count) then
          txt := FPathsShow[n];
        R.Right := R.Left + FBmp.Canvas.TextWidth(txt) + 9;
      end;
      if done then
        ShowArrow := (FShowArrow) or (FPathOff = FPathsShow.Count);
      if ShowArrow then
        if (FState <> spNormal) and ((n) = FHotPos) then
        begin
          if Assigned(FOnOwnerDraw) then
            FOnOwnerDraw(Self, FBmp.Canvas, R, txt, FState, not bool)
          else
            PaintButton(FBmp.Canvas, R, txt, FState, not bool);
        end
        else
        begin
          if Assigned(FOnOwnerDraw) then
            FOnOwnerDraw(Self, FBmp.Canvas, R, txt, spNormal, not bool)
          else
            PaintButton(FBmp.Canvas, R, txt, spNormal, not bool);
        end;
      R.Left := R.Right;
      bool := not bool;
    end;
  end;
  BitBlt(Canvas.Handle, 0, 0, FBmp.Width, FBmp.Height, FBmp.Canvas.Handle, 0, 0,
    SRCCOPY);
end;

procedure TrkSmartPath.WMERASEBKGND(var Message: TWMEraseBkgnd);
begin
  Message.Result := 1;
end;

procedure TrkSmartPath.WMPaint(var Message: TWMPaint);
begin
  PaintHandler(Message);
end;

procedure TrkSmartPath.CMCtl3DChanged(var Message: TMessage);
begin
  if NewStyleControls and (FBorderStyle = bsSingle) then
    RecreateWnd;
  inherited;
end;

procedure TrkSmartPath.WMWindowPosChanged(var Message: TWMWindowPosChanged);
begin
  inherited;
  if not (csLoading in ComponentState) then
    Resize;
end;

procedure TrkSmartPath.CMRefresh(var Message: TMessage);
begin
  if (csDestroying in ComponentState) then
    Exit;
  FHotPos := -1;
  FPathIdx := -1;
  FPopupPath := '';
  FPopUp := False;
  FClicked := False;
  FHotPt := Point(-1, -1);
  FState := spNormal;
  Invalidate;
  inherited;
end;

procedure TrkSmartPath.SetBorderColor(const Value: TColor);
begin
  FBorderColor := Value;
  Invalidate;
end;

procedure TrkSmartPath.SetBorderStyle(Value: TBorderStyle);
begin
  if Value <> FBorderStyle then
  begin
    FBorderStyle := Value;
    RecreateWnd;
  end;
end;

procedure TrkSmartPath.SetFont(Value: TFont);
begin
  if (FFont <> Value) then
    FFont.Assign(Value);
  FPathEdit.Font.Assign(Value);
  Invalidate;
end;

procedure TrkSmartPath.SetFramed(const Value: Boolean);
begin
  FFramed := Value;
  Invalidate;
end;

procedure TrkSmartPath.SetPath(Value: string);
var
  SFI: TSHFileInfo;
  SR: TSearchRec;
  IsFound: Boolean;
  i: Integer;
  s, p: string;
  SelectedPIDL: PItemIDList;
begin
  if Value = '' then
  begin
    ComputerAsDefault:= FDefault;
    FPath := '';
    FPaths.Clear;
    FPaths.Add(FEmptyPathText);
    FPathsShow.Clear;
    FPathsShow.Add(FEmptyPathText);
    FImageIdx := FEmptyPathIcon;
    FPathOff := 0;
    FShowArrow := False;
  end
  else
  begin
    if (FPath = Value) or not (DirectoryExists(Value)) then
      Exit;
    FPath := AddBackSlash(Value);
    FPaths.Clear;
    FPaths.Text := Trim(StringReplace(FPath, '\', #13, [rfReplaceAll]));

    SelectedPIDL := PathToPIDL(FPath);
    SHGetFileInfo(PChar(SelectedPIDL), 0, SFI, SizeOf(SFI), SHGFI_PIDL or
      SHGFI_SYSICONINDEX or SHGFI_DISPLAYNAME or SHGFI_TYPENAME);

    FImageIdx := SFI.iIcon;
    FClicked := False;
    FShowArrow := False;

    s := '';
    FPathsShow.Clear;
    for i := 0 to FPaths.Count - 1 do
    begin
      s := s + FPaths[i] + '\';
      ShGetFileInfo(PChar(s), 0, SFI, SizeOf(TSHFileInfo), SHGFI_SYSICONINDEX
        or SHGFI_DISPLAYNAME or SHGFI_TYPENAME or SHGFI_SMALLICON);
      p := SFI.szDisplayName;
      if p = '' then
        p := FPaths[i];
      FPathsShow.Add(SFI.szDisplayName);
    end;

    // Check if folder got folder items...
    IsFound := FindFirst(Value + '\*.*', faAnyFile - faHidden, SR) = 0;
    while IsFound do
    begin
      if (SR.Name <> '.') and (SR.Name <> '..') then
        FShowArrow := ((SR.Attr and faDirectory) <> 0);
      IsFound := (FindNext(SR) = 0);
      if FShowArrow then
        IsFound := False;
    end;
    FindClose(SR);
    FPathOff := CalcPath;
  end;
  Invalidate;
  if Assigned(FOnPathChanged) then
    FOnPathChanged(Self);
end;

procedure TrkSmartPath.SetColor(const Value: TColor);
begin
  FColor := Value;
  Invalidate
end;

procedure TrkSmartPath.SetCustomPaths(Value: TStringList);
begin
  FCustomPaths.Assign(Value);
end;

procedure TrkSmartPath.SetDefault(const Value: Boolean);
var
  PIDL: PItemIDList;
  FileInfo: TSHFileInfo;
begin
  FDefault := Value;
  if Value = True then
    SHGetSpecialFolderLocation(Application.Handle, CSIDL_DRIVES, PIDL)
  else
    Exit;
  if SHGetFileInfo(PChar(PIDL), 0, FileInfo, SizeOf(TSHFileInfo),
    SHGFI_PIDL or SHGFI_SYSICONINDEX or SHGFI_DISPLAYNAME) <> 0 then
  begin
    EmptyPathIcon := FileInfo.iIcon;
    EmptyPathText := FileInfo.szDisplayName;
  end;
  Invalidate;
end;

procedure TrkSmartPath.SetPathsToAdd(Value: TspPaths);
begin
  FSpecialFolders := Value;
end;

// Calculate how many parts of the path that will fit in view
function TrkSmartPath.CalcPath: Integer;
var
  i, w: Integer;
begin
  Result := 0;
  if (Path = '') then
    Exit;
  w := 22 + 14;
  for i := 0 to FPathsShow.Count - 1 do
    w := w + FBmp.Canvas.TextWidth(FPathsShow[i]) + 23;
  if not FShowArrow then
    w := w - 14;
  if w < ClientWidth - 8 then
    Exit;
  w := w - (ClientWidth - 8);
  i := 0;
  FBmp.Canvas.Font.Assign(FFont);
  while (w > 0) and (i < FPathsShow.Count) do
  begin
    w := w - (FBmp.Canvas.TextWidth(FPathsShow[i]) + 23);
    i := i + 1;
  end;
  if w > 0 then
    Result := FPaths.Count
  else
    Result := i;
end;

constructor TrkSmartPath.Create(AOwner: TComponent);
var
  TmpHandle: THandle;
begin
  inherited;
  if AOwner is TWinControl then
    Parent := TWinControl(AOwner);
  Width := 150;
  Height := 25;
  ControlStyle := ControlStyle + [csOpaque, csReplicatable, csNeedsBorderPaint];
  FState := spNormal;
  FClicked := False;
  FEmptyPathicon := -1;
  FEmptyPathText := '';
  FPopupPath := '';
  FPath := '';
  FPaths := TStringList.Create;
  FPathsShow := TStringList.Create;
  FCustomPaths := TStringList.Create;
  FMenuLookup := TStringList.Create;
  FPathOff := -1;
  FImageIdx := -1;
  FFont := TFont.Create;
  FFont.Name := 'Tahoma';
  FFont.Color := clBlack;
  FFont.Size := 8;
  FBmp := TBitmap.Create;
  FBmp.PixelFormat := pf24Bit;
  FBorderStyle := bsNone;
  FBorderColor := clSilver;
  FColor := clWindow;
  FPathIdx := -1;
  FFramed := True;
  FAuto := True;
  FAllowEdit:= False;
  FAllowSearch:= False;
  FSpecialFolders := [spDocuments, spDesktop];
  FImages := TImagelist.Create(nil);
  FImages.ShareImages := True;
  FImages.DrawingStyle := dsTransparent;
  FDesktopFolder := GetSpecialFolderPath(CSIDL_PERSONAL);
  TmpHandle := SysImageListHandle(FDesktopFolder, False);
  if TmpHandle <> 0 then
    FImages.Handle := TmpHandle;
  FMenu := TPopUpMenu.Create(nil);
  FMenu.OnPopup := FMenuPopup;
  FMenu.Images := FImages;
  FPathEdit := TEdit.Create(nil);
  FPathEdit.Parent := Self;
  FPathEdit.ParentFont := False;
  FPathEdit.BorderStyle := bsNone;
  FPathEdit.AutoSize := False;
  FPathEdit.Ctl3D := True;
  FPathEdit.Visible := False;
  FPathEdit.OnKeyDown := EditKeyDown;
  FPathEdit.OnKeyPress := EditKeyPress;
  FPathEdit.OnExit := EditExit;
  FNewFolder := 'NewFolder';
  Enabled := True;
  Visible := True;
  ComputerAsDefault:= True;
  Path := FDesktopFolder;
end;

destructor TrkSmartPath.Destroy;
begin
  FPathEdit.Free;
  FMenu.Free;
  FMenuLookup.Free;
  FImages.Free;
  FCustomPaths.Free;
  FPaths.Free;
  FPathsShow.Free;
  FFont.Free;
  FBmp.Free;
  inherited Destroy;
end;

procedure TrkSmartPath.Resize;
begin
  FPathOff := CalcPath;
  Invalidate;
end;

procedure TrkSmartPath.CreateParams(var Params: TCreateParams);
const
  BorderStyles: array[TBorderStyle] of DWord = (0, WS_BORDER);
begin
  inherited CreateParams(Params);
  with Params do
  begin
    Style := Style or BorderStyles[FBorderStyle];
    if NewStyleControls and Ctl3D and (FBorderStyle = bsSingle) then
    begin
      Style := Style and not WS_BORDER;
      ExStyle := ExStyle or WS_EX_CLIENTEDGE;
    end;
  end;
end;

procedure TrkSmartPath.MouseDown(Button: TMouseButton; Shift: TShiftState; X,
  Y: Integer);
begin
  if (Button = mbLeft) and (not FPathEdit.Visible) then
  begin
    FHotPt := Point(X, Y);
    CalcState;
    if (FState = spArrowHot) then
      FPopUp := True
    else
      FPopUp := False;
    FClicked := True;
    PaintPath;
    if FAuto then
    begin
      if (FPopUp) and (FPopupPath <> '') then
      begin
        FMenu.Popup(FPopupPos.X, FPopupPos.Y);
        PostMessage(Handle, CM_REFRESH, 0, 0);
      end;
    end
    else
    begin
      if (FPopUp) and Assigned(FCustomPopup) and (FPopupPath <> '') then
      begin
        FCustomPopup.Popup(FPopupPos.X, FPopupPos.Y);
        PostMessage(Handle, CM_REFRESH, 0, 0);
      end;
    end;
    if (FState = spNormal) and (FAllowEdit = True) then
      DoEdit(False);
  end;
  if (FState = spNormal) then
    FPopupPath := Path;
  inherited;
end;

procedure TrkSmartPath.MouseUp(Button: TMouseButton; Shift: TShiftState; X,
  Y: Integer);
begin
  if FPopUp then
    Exit;
  FClicked := False;
  FPopUp := False;
  Invalidate;
  inherited;
end;

procedure TrkSmartPath.MouseMove(Shift: TShiftState; X, Y: Integer);
begin
  if FPathEdit.Visible then
    Exit;
  FHotPt := Point(X, Y);
  Invalidate;
end;

procedure TrkSmartPath.CMMouseLeave(var Message: TMessage);
begin
  if FPopUp then
    Exit;
  FClicked := False;
  FPopUp := False;
  FHotPt := Point(-1, -1);
  Invalidate;
end;

procedure Register;
begin
  RegisterComponents('rmklever', [TrkSmartPath]);
end;

end.

